from __future__ import absolute_import

import datetime

from graphql.language import ast

from .scalars import Scalar

try:
    import iso8601
except:
    raise ImportError(
        "iso8601 package is required for DateTime Scalar.\n"
        "You can install it using: pip install iso8601."
    )


class DateTime(Scalar):
    '''
    The `DateTime` scalar type represents a DateTime
    value as specified by
    [iso8601](https://en.wikipedia.org/wiki/ISO_8601).
    '''

    @staticmethod
    def serialize(dt):
        assert isinstance(dt, (datetime.datetime, datetime.date)), (
            'Received not compatible datetime "{}"'.format(repr(dt))
        )
        return dt.isoformat()

    @staticmethod
    def parse_literal(node):
        if isinstance(node, ast.StringValue):
            return iso8601.parse_date(node.value)

    @staticmethod
    def parse_value(value):
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")



from ..utils.orderedtype import OrderedType


class UnmountedType(OrderedType):
    '''
    This class acts a proxy for a Graphene Type, so it can be mounted
    dynamically as Field, InputField or Argument.

    Instead of writing
    >>> class MyObjectType(ObjectType):
    >>>     my_field = Field(String(), description='Description here')

    It let you write
    >>> class MyObjectType(ObjectType):
    >>>     my_field = String(description='Description here')
    '''

    def __init__(self, *args, **kwargs):
        super(UnmountedType, self).__init__()
        self.args = args
        self.kwargs = kwargs

    def get_type(self):
        raise NotImplementedError("get_type not implemented in {}".format(self))

    def Field(self):  # noqa: N802
        '''
        Mount the UnmountedType as Field
        '''
        from .field import Field
        return Field(
            self.get_type(),
            *self.args,
            _creation_counter=self.creation_counter,
            **self.kwargs
        )

    def InputField(self):  # noqa: N802
        '''
        Mount the UnmountedType as InputField
        '''
        from .inputfield import InputField
        return InputField(
            self.get_type(),
            *self.args,
            _creation_counter=self.creation_counter,
            **self.kwargs
        )

    def Argument(self):  # noqa: N802
        '''
        Mount the UnmountedType as Argument
        '''
        from .argument import Argument
        return Argument(
            self.get_type(),
            *self.args,
            _creation_counter=self.creation_counter,
            **self.kwargs
        )

    def __eq__(self, other):
        return (
            self is other or (
                isinstance(other, UnmountedType) and
                self.get_type() == other.get_type() and
                self.args == other.args and
                self.kwargs == other.kwargs
            )
        )



import six

from ..utils.is_base_type import is_base_type
from .options import Options


class UnionMeta(type):

    def __new__(cls, name, bases, attrs):
        # Also ensure initialization is only performed for subclasses of
        # Union
        if not is_base_type(bases, UnionMeta):
            return type.__new__(cls, name, bases, attrs)

        options = Options(
            attrs.pop('Meta', None),
            name=name,
            description=attrs.get('__doc__'),
            types=(),
        )

        assert (
            isinstance(options.types, (list, tuple)) and
            len(options.types) > 0
        ), 'Must provide types for Union {}.'.format(options.name)

        return type.__new__(cls, name, bases, dict(attrs, _meta=options))

    def __str__(cls):  # noqa: N805
        return cls._meta.name


class Union(six.with_metaclass(UnionMeta)):
    '''
    Union Type Definition

    When a field can return one of a heterogeneous set of types, a Union type
    is used to describe what types are possible as well as providing a function
    to determine which type is actually used when the field is resolved.
    '''

    resolve_type = None

    def __init__(self, *args, **kwargs):
        raise Exception("An Union cannot be intitialized")



from collections import OrderedDict

import six

from ..utils.is_base_type import is_base_type
from .options import Options
from .unmountedtype import UnmountedType

try:
    from enum import Enum as PyEnum
except ImportError:
    from ..pyutils.enum import Enum as PyEnum


class EnumTypeMeta(type):

    def __new__(cls, name, bases, attrs):
        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        if not is_base_type(bases, EnumTypeMeta):
            return type.__new__(cls, name, bases, attrs)

        options = Options(
            attrs.pop('Meta', None),
            name=name,
            description=attrs.get('__doc__'),
            enum=None,
        )
        if not options.enum:
            options.enum = PyEnum(cls.__name__, attrs)

        new_attrs = OrderedDict(attrs, _meta=options, **options.enum.__members__)
        return type.__new__(cls, name, bases, new_attrs)

    def __prepare__(name, bases, **kwargs):  # noqa: N805
        return OrderedDict()

    def __call__(cls, *args, **kwargs):  # noqa: N805
        if cls is Enum:
            description = kwargs.pop('description', None)
            return cls.from_enum(PyEnum(*args, **kwargs), description=description)
        return super(EnumTypeMeta, cls).__call__(*args, **kwargs)

    def from_enum(cls, enum, description=None):  # noqa: N805
        meta_class = type('Meta', (object,), {'enum': enum, 'description': description})
        return type(meta_class.enum.__name__, (Enum,), {'Meta': meta_class})

    def __str__(cls):  # noqa: N805
        return cls._meta.name


class Enum(six.with_metaclass(EnumTypeMeta, UnmountedType)):
    '''
    Enum Type Definition

    Some leaf values of requests and input values are Enums. GraphQL serializes
    Enum values as strings, however internally Enums can be represented by any
    kind of type, often integers.
    '''

    def get_type(self):
        return type(self)



from .unmountedtype import UnmountedType


class Structure(UnmountedType):
    '''
    A structure is a GraphQL type instance that
    wraps a main type with certain structure.
    '''

    def __init__(self, of_type, *args, **kwargs):
        super(Structure, self).__init__(*args, **kwargs)
        self.of_type = of_type

    def get_type(self):
        return self


class List(Structure):
    '''
    List Modifier

    A list is a kind of type marker, a wrapping type which points to another
    type. Lists are often created within the context of defining the fields of
    an object type.
    '''

    def __str__(self):
        return '[{}]'.format(self.of_type)

    def __eq__(self, other):
        return isinstance(other, List) and (
            self.of_type == other.of_type and
            self.args == other.args and
            self.kwargs == other.kwargs
        )


class NonNull(Structure):
    '''
    Non-Null Modifier

    A non-null is a kind of type marker, a wrapping type which points to another
    type. Non-null types enforce that their values are never null and can ensure
    an error is raised if this ever occurs during a request. It is useful for
    fields which you can make a strong guarantee on non-nullability, for example
    usually the id field of a database row will never be null.

    Note: the enforcement of non-nullability occurs within the executor.
    '''

    def __init__(self, *args, **kwargs):
        super(NonNull, self).__init__(*args, **kwargs)
        assert not isinstance(self.of_type, NonNull), (
            'Can only create NonNull of a Nullable GraphQLType but got: {}.'
        ).format(type)

    def __str__(self):
        return '{}!'.format(self.of_type)

    def __eq__(self, other):
        return isinstance(other, NonNull) and (
            self.of_type == other.of_type and
            self.args == other.args and
            self.kwargs == other.kwargs
        )



from .node import Node, is_node, GlobalID
from .mutation import ClientIDMutation
from .connection import Connection, ConnectionField, PageInfo

__all__ = [
    'Node',
    'is_node',
    'GlobalID',
    'ClientIDMutation',
    'Connection',
    'ConnectionField',
    'PageInfo',
]



from functools import total_ordering


@total_ordering
class OrderedType(object):
    creation_counter = 1

    def __init__(self, _creation_counter=None):
        self.creation_counter = _creation_counter or self.gen_counter()

    @staticmethod
    def gen_counter():
        counter = OrderedType.creation_counter
        OrderedType.creation_counter += 1
        return counter

    def reset_counter(self):
        self.creation_counter = self.gen_counter()

    def __eq__(self, other):
        # Needed for @total_ordering
        if isinstance(self, type(other)):
            return self.creation_counter == other.creation_counter
        return NotImplemented

    def __lt__(self, other):
        # This is needed because bisect does not take a comparison function.
        if isinstance(other, OrderedType):
            return self.creation_counter < other.creation_counter
        return NotImplemented

    def __gt__(self, other):
        # This is needed because bisect does not take a comparison function.
        if isinstance(other, OrderedType):
            return self.creation_counter > other.creation_counter
        return NotImplemented

    def __hash__(self):
        return hash((self.creation_counter))



from functools import partial

import six

from graphql_relay import from_global_id, to_global_id

from ..types import ID, Field, Interface, ObjectType
from ..types.interface import InterfaceMeta


def is_node(objecttype):
    '''
    Check if the given objecttype has Node as an interface
    '''
    assert issubclass(objecttype, ObjectType), (
        'Only ObjectTypes can have a Node interface.'
    )
    for i in objecttype._meta.interfaces:
        if issubclass(i, Node):
            return True
    return False


def get_default_connection(cls):
    from .connection import Connection
    assert issubclass(cls, ObjectType), (
        'Can only get connection type on implemented Nodes.'
    )

    class Meta:
        node = cls

    return type('{}Connection'.format(cls.__name__), (Connection,), {'Meta': Meta})


class GlobalID(Field):

    def __init__(self, node, *args, **kwargs):
        super(GlobalID, self).__init__(ID, *args, **kwargs)
        self.node = node

    @staticmethod
    def id_resolver(parent_resolver, node, root, args, context, info):
        id = parent_resolver(root, args, context, info)
        return node.to_global_id(info.parent_type.name, id)  # root._meta.name

    def get_resolver(self, parent_resolver):
        return partial(self.id_resolver, parent_resolver, self.node)


class NodeMeta(InterfaceMeta):

    def __new__(cls, name, bases, attrs):
        cls = InterfaceMeta.__new__(cls, name, bases, attrs)
        cls._meta.fields['id'] = GlobalID(cls, required=True, description='The ID of the object.')
        return cls


class NodeField(Field):

    def __init__(self, node, type=False, deprecation_reason=None,
                 name=None, **kwargs):
        assert issubclass(node, Node), 'NodeField can only operate in Nodes'
        type = type or node
        super(NodeField, self).__init__(
            type,
            description='The ID of the object',
            id=ID(required=True),
            resolver=node.node_resolver
        )


class Node(six.with_metaclass(NodeMeta, Interface)):
    '''An object with an ID'''

    @classmethod
    def Field(cls, *args, **kwargs):  # noqa: N802
        return NodeField(cls, *args, **kwargs)

    @classmethod
    def node_resolver(cls, root, args, context, info):
        return cls.get_node_from_global_id(args.get('id'), context, info)

    @classmethod
    def get_node_from_global_id(cls, global_id, context, info):
        try:
            _type, _id = cls.from_global_id(global_id)
            graphene_type = info.schema.get_type(_type).graphene_type
            # We make sure the ObjectType implements the "Node" interface
            assert cls in graphene_type._meta.interfaces
        except:
            return None
        get_node = getattr(graphene_type, 'get_node', None)
        if get_node:
            return get_node(_id, context, info)

    @classmethod
    def from_global_id(cls, global_id):
        return from_global_id(global_id)

    @classmethod
    def to_global_id(cls, type, id):
        return to_global_id(type, id)

    @classmethod
    def implements(cls, objecttype):
        get_connection = getattr(objecttype, 'get_connection', None)
        if not get_connection:
            get_connection = partial(get_default_connection, objecttype)

        objecttype.Connection = get_connection()



from collections import OrderedDict

from .dynamic import Dynamic
from .field import Field
from .inputfield import InputField
from .unmountedtype import UnmountedType


def merge(*dicts):
    '''
    Merge the dicts into one
    '''
    merged = OrderedDict()
    for _dict in dicts:
        merged.update(_dict)
    return merged


def get_base_fields(bases, _as=None):
    '''
    Get all the fields in the given bases
    '''
    fields = OrderedDict()
    from ..types import AbstractType, Interface
    # We allow inheritance in AbstractTypes and Interfaces but not ObjectTypes
    inherited_bases = (AbstractType, Interface)
    for base in bases:
        if base in inherited_bases or not issubclass(base, inherited_bases):
            continue
        for name, field in base._meta.fields.items():
            if name in fields:
                continue
            fields[name] = get_field_as(field, _as=_as)

    return fields


def mount_as(unmounted_field, _as):
    '''
    Mount the UnmountedType dinamically as Field or InputField
    '''
    if _as is None:
        return unmounted_field

    elif _as is Field:
        return unmounted_field.Field()

    elif _as is InputField:
        return unmounted_field.InputField()

    raise Exception(
        'Unmounted field "{}" cannot be mounted in {}.'.format(
            unmounted_field, _as
        )
    )


def get_field_as(value, _as=None):
    '''
    Get type mounted
    '''
    if isinstance(value, (Field, InputField, Dynamic)):
        return value
    elif isinstance(value, UnmountedType):
        return mount_as(value, _as)


def yank_fields_from_attrs(attrs, _as=None, delete=True, sort=True):
    '''
    Extract all the fields in given attributes (dict)
    and return them ordered
    '''
    fields_with_names = []
    for attname, value in list(attrs.items()):
        field = get_field_as(value, _as)
        if not field:
            continue
        fields_with_names.append((attname, field))
        if delete:
            del attrs[attname]

    if sort:
        fields_with_names = sorted(fields_with_names, key=lambda f: f[1])
    return OrderedDict(fields_with_names)



# flake8: noqa

from .objecttype import ObjectType
from .abstracttype import AbstractType
from .interface import Interface
from .mutation import Mutation
from .scalars import Scalar, String, ID, Int, Float, Boolean
from .schema import Schema
from .structures import List, NonNull
from .enum import Enum
from .field import Field
from .inputfield import InputField
from .argument import Argument
from .inputobjecttype import InputObjectType
from .dynamic import Dynamic
from .union import Union


__all__ = [
    'AbstractType',
    'ObjectType',
    'InputObjectType',
    'Interface',
    'Mutation',
    'Enum',
    'Field',
    'InputField',
    'Schema',
    'Scalar',
    'String',
    'ID',
    'Int',
    'Float',
    'Boolean',
    'List',
    'NonNull',
    'Argument',
    'Dynamic',
    'Union',
]



import inspect

from ..utils.orderedtype import OrderedType


class Dynamic(OrderedType):
    '''
    A Dynamic Type let us get the type in runtime when we generate
    the schema. So we can have lazy fields.
    '''

    def __init__(self, type, _creation_counter=None):
        super(Dynamic, self).__init__(_creation_counter=_creation_counter)
        assert inspect.isfunction(type)
        self.type = type

    def get_type(self):
        return self.type()



import six

from graphql.language.ast import (BooleanValue, FloatValue, IntValue,
                                  StringValue)

from ..utils.is_base_type import is_base_type
from .options import Options
from .unmountedtype import UnmountedType


class ScalarTypeMeta(type):

    def __new__(cls, name, bases, attrs):
        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        if not is_base_type(bases, ScalarTypeMeta):
            return type.__new__(cls, name, bases, attrs)

        options = Options(
            attrs.pop('Meta', None),
            name=name,
            description=attrs.get('__doc__'),
        )

        return type.__new__(cls, name, bases, dict(attrs, _meta=options))

    def __str__(cls):  # noqa: N802
        return cls._meta.name


class Scalar(six.with_metaclass(ScalarTypeMeta, UnmountedType)):
    '''
    Scalar Type Definition

    The leaf values of any request and input values to arguments are
    Scalars (or Enums) and are defined with a name and a series of functions
    used to parse input from ast or variables and to ensure validity.
    '''

    serialize = None
    parse_value = None
    parse_literal = None

    @classmethod
    def get_type(cls):
        return cls

# As per the GraphQL Spec, Integers are only treated as valid when a valid
# 32-bit signed integer, providing the broadest support across platforms.
#
# n.b. JavaScript's integers are safe between -(2^53 - 1) and 2^53 - 1 because
# they are internally represented as IEEE 754 doubles.
MAX_INT = 2147483647
MIN_INT = -2147483648


class Int(Scalar):
    '''
    The `Int` scalar type represents non-fractional signed whole numeric
    values. Int can represent values between -(2^53 - 1) and 2^53 - 1 since
    represented in JSON as double-precision floating point numbers specified
    by [IEEE 754](http://en.wikipedia.org/wiki/IEEE_floating_point).
    '''

    @staticmethod
    def coerce_int(value):
        try:
            num = int(value)
        except ValueError:
            try:
                num = int(float(value))
            except ValueError:
                return None
        if MIN_INT <= num <= MAX_INT:
            return num

    serialize = coerce_int
    parse_value = coerce_int

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, IntValue):
            num = int(ast.value)
            if MIN_INT <= num <= MAX_INT:
                return num


class Float(Scalar):
    '''
    The `Float` scalar type represents signed double-precision fractional
    values as specified by
    [IEEE 754](http://en.wikipedia.org/wiki/IEEE_floating_point).
    '''

    @staticmethod
    def coerce_float(value):
        try:
            return float(value)
        except ValueError:
            return None

    serialize = coerce_float
    parse_value = coerce_float

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, (FloatValue, IntValue)):
            return float(ast.value)


class String(Scalar):
    '''
    The `String` scalar type represents textual data, represented as UTF-8
    character sequences. The String type is most often used by GraphQL to
    represent free-form human-readable text.
    '''

    @staticmethod
    def coerce_string(value):
        if isinstance(value, bool):
            return u'true' if value else u'false'
        return six.text_type(value)

    serialize = coerce_string
    parse_value = coerce_string

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, StringValue):
            return ast.value


class Boolean(Scalar):
    '''
    The `Boolean` scalar type represents `true` or `false`.
    '''

    serialize = bool
    parse_value = bool

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, BooleanValue):
            return ast.value


class ID(Scalar):
    '''
    The `ID` scalar type represents a unique identifier, often used to
    refetch an object or as key for a cache. The ID type appears in a JSON
    response as a String; however, it is not intended to be human-readable.
    When expected as an input type, any string (such as `"4"`) or integer
    (such as `4`) input value will be accepted as an ID.
    '''

    serialize = str
    parse_value = str

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, (StringValue, IntValue)):
            return ast.value
